import json
import logging
import secrets
import shutil
import subprocess
import sys
from pathlib import Path

log = logging.getLogger(__name__)


def _update_toc_linux(docx_file: Path) -> None:
    """TOC bindings for linux"""
    # This method hangs if item is already open, so we cheat a little here
    tmp_file = str(docx_file) + f".{secrets.token_hex(4)}.docx"
    tmp_file = Path(tmp_file)
    shutil.copy(docx_file, tmp_file)

    # Source: https://github.com/python-openxml/python-docx/issues/1207#issuecomment-1924053420
    subprocess.call(
        [
            "libreoffice",
            "--headless",
            f"macro:///Standard.Module1.UpdateTOC({str(tmp_file)})",
        ],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )

    shutil.copy(tmp_file, docx_file)
    tmp_file.unlink()


def _create_pdf_windows(docx_file: Path) -> None:
    import win32com.client

    word = win32com.client.Dispatch("Word.Application")
    wdFormatPDF = 17

    docx_filepath = docx_file
    pdf_filepath = Path(f"{docx_file.stem}.pdf").resolve()
    doc = word.Documents.Open(str(docx_filepath))
    try:
        doc.SaveAs(str(pdf_filepath), FileFormat=wdFormatPDF)
    except:
        raise
    finally:
        doc.Close(0)

    word.Quit()


def _create_pdf_linux(docx_file: Path) -> None:
    subprocess.call(
        [
            "libreoffice",
            "--convert-to",
            "pdf",
            str(docx_file),
        ],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )


def _create_pdf_macos(docx_file: Path) -> None:
    log.warning("DOCX -> PDF on mac is untested. Any issues please raise an issue.")
    script = (Path(__file__).parent / "convert.jxa").resolve()
    cmd = [
        "/usr/bin/osascript",
        "-l",
        "JavaScript",
        str(script),
        str(docx_file),
        str(Path(f"{docx_file.stem}.pdf").resolve()),
    ]

    process = subprocess.Popen(cmd, stderr=subprocess.PIPE)
    process.wait()
    if process.returncode != 0:
        msg = process.stderr.read().decode().rstrip()
        if "application can't be found" in msg.lower():
            raise EnvironmentError("Microsoft Word is not available.")
        raise RuntimeError(msg)

    def stderr_results(process):
        while True:
            output_line = process.stderr.readline().rstrip()
            if not output_line:
                break
            yield output_line.decode("utf-8")

    for line in stderr_results(process):
        try:
            msg = json.loads(line)
        except ValueError:
            continue
        if msg["result"] == "error":
            print(msg)
            sys.exit(1)


def export_libre_macro(
    macro_folder: Path = Path("~/.config/libreoffice/4/user/basic/Standard"),
) -> None:
    """Automatically moves the LibreOffice macro file to `macro_folder`.

    Warning, this overrides Module1.xba

    :py:class:`Path` is where your macros live
    """
    macro_folder = macro_folder.expanduser()
    module_file = Path(__file__).parent.resolve() / "Module1.xba"
    shutil.copy(module_file, macro_folder)


def update_toc(docx_file: Path | str) -> None:
    """Update a TOC within a word document.

    If you are on linux, please call `export_libre_macro` first.
    """
    if isinstance(docx_file, str):
        docx_file = Path(docx_file)

    docx_file = docx_file.resolve().absolute()

    if sys.platform == "linux":
        _update_toc_linux(docx_file)
    elif sys.platform == "win32":
        raise ValueError("Windows is not yet implemented yet.")
    else:
        raise ValueError(f"{sys.platform} is not implemented")


def document_to_pdf(docx_file: Path | str) -> None:
    """Create a PDF from a word document.

    Consider calling the relevant API's yourself
    if you need to add extra context to calls
    such as watermark arguments.
    """
    if isinstance(docx_file, str):
        docx_file = Path(docx_file)

    docx_file = docx_file.absolute()

    if sys.platform == "linux":
        _create_pdf_linux(docx_file)
    elif sys.platform == "win32":
        _create_pdf_windows(docx_file)
    elif sys.platform == "darwin":
        _create_pdf_macos(docx_file)
    else:
        raise ValueError(f"{sys.platform} is not implemented")



"""DrawingML-related objects are in this subpackage."""

from __future__ import annotations

from typing import TYPE_CHECKING

from docx.oxml.drawing import CT_Drawing
from docx.shared import Parented

if TYPE_CHECKING:
    import docx.types as t


class Drawing(Parented):
    """Container for a DrawingML object."""

    def __init__(self, drawing: CT_Drawing, parent: t.ProvidesStoryPart):
        super().__init__(parent)
        self._parent = parent
        self._drawing = self._element = drawing



from struct import Struct

from .constants import MIME_TYPE
from .image import BaseImageHeader


class Gif(BaseImageHeader):
    """Image header parser for GIF images.

    Note that the GIF format does not support resolution (DPI) information. Both
    horizontal and vertical DPI default to 72.
    """

    @classmethod
    def from_stream(cls, stream):
        """Return |Gif| instance having header properties parsed from GIF image in
        `stream`."""
        px_width, px_height = cls._dimensions_from_stream(stream)
        return cls(px_width, px_height, 72, 72)

    @property
    def content_type(self):
        """MIME content type for this image, unconditionally `image/gif` for GIF
        images."""
        return MIME_TYPE.GIF

    @property
    def default_ext(self):
        """Default filename extension, always 'gif' for GIF images."""
        return "gif"

    @classmethod
    def _dimensions_from_stream(cls, stream):
        stream.seek(6)
        bytes_ = stream.read(4)
        struct = Struct("<HH")
        px_width, px_height = struct.unpack(bytes_)
        return px_width, px_height



"""Low-level, read-only API to a serialized Open Packaging Convention (OPC) package."""

from docx.opc.constants import RELATIONSHIP_TARGET_MODE as RTM
from docx.opc.oxml import parse_xml
from docx.opc.packuri import PACKAGE_URI, PackURI
from docx.opc.phys_pkg import PhysPkgReader
from docx.opc.shared import CaseInsensitiveDict


class PackageReader:
    """Provides access to the contents of a zip-format OPC package via its
    :attr:`serialized_parts` and :attr:`pkg_srels` attributes."""

    def __init__(self, content_types, pkg_srels, sparts):
        super(PackageReader, self).__init__()
        self._pkg_srels = pkg_srels
        self._sparts = sparts

    @staticmethod
    def from_file(pkg_file):
        """Return a |PackageReader| instance loaded with contents of `pkg_file`."""
        phys_reader = PhysPkgReader(pkg_file)
        content_types = _ContentTypeMap.from_xml(phys_reader.content_types_xml)
        pkg_srels = PackageReader._srels_for(phys_reader, PACKAGE_URI)
        sparts = PackageReader._load_serialized_parts(phys_reader, pkg_srels, content_types)
        phys_reader.close()
        return PackageReader(content_types, pkg_srels, sparts)

    def iter_sparts(self):
        """Generate a 4-tuple `(partname, content_type, reltype, blob)` for each of the
        serialized parts in the package."""
        for s in self._sparts:
            yield (s.partname, s.content_type, s.reltype, s.blob)

    def iter_srels(self):
        """Generate a 2-tuple `(source_uri, srel)` for each of the relationships in the
        package."""
        for srel in self._pkg_srels:
            yield (PACKAGE_URI, srel)
        for spart in self._sparts:
            for srel in spart.srels:
                yield (spart.partname, srel)

    @staticmethod
    def _load_serialized_parts(phys_reader, pkg_srels, content_types):
        """Return a list of |_SerializedPart| instances corresponding to the parts in
        `phys_reader` accessible by walking the relationship graph starting with
        `pkg_srels`."""
        sparts = []
        part_walker = PackageReader._walk_phys_parts(phys_reader, pkg_srels)
        for partname, blob, reltype, srels in part_walker:
            content_type = content_types[partname]
            spart = _SerializedPart(partname, content_type, reltype, blob, srels)
            sparts.append(spart)
        return tuple(sparts)

    @staticmethod
    def _srels_for(phys_reader, source_uri):
        """Return |_SerializedRelationships| instance populated with relationships for
        source identified by `source_uri`."""
        rels_xml = phys_reader.rels_xml_for(source_uri)
        return _SerializedRelationships.load_from_xml(source_uri.baseURI, rels_xml)

    @staticmethod
    def _walk_phys_parts(phys_reader, srels, visited_partnames=None):
        """Generate a 4-tuple `(partname, blob, reltype, srels)` for each of the parts
        in `phys_reader` by walking the relationship graph rooted at srels."""
        if visited_partnames is None:
            visited_partnames = []
        for srel in srels:
            if srel.is_external:
                continue
            partname = srel.target_partname
            if partname in visited_partnames:
                continue
            visited_partnames.append(partname)
            reltype = srel.reltype
            part_srels = PackageReader._srels_for(phys_reader, partname)
            blob = phys_reader.blob_for(partname)
            yield (partname, blob, reltype, part_srels)
            next_walker = PackageReader._walk_phys_parts(phys_reader, part_srels, visited_partnames)
            for partname, blob, reltype, srels in next_walker:
                yield (partname, blob, reltype, srels)


class _ContentTypeMap:
    """Value type providing dictionary semantics for looking up content type by part
    name, e.g. ``content_type = cti['/ppt/presentation.xml']``."""

    def __init__(self):
        super(_ContentTypeMap, self).__init__()
        self._overrides = CaseInsensitiveDict()
        self._defaults = CaseInsensitiveDict()

    def __getitem__(self, partname):
        """Return content type for part identified by `partname`."""
        if not isinstance(partname, PackURI):
            tmpl = "_ContentTypeMap key must be <type 'PackURI'>, got %s"
            raise KeyError(tmpl % type(partname))
        if partname in self._overrides:
            return self._overrides[partname]
        if partname.ext in self._defaults:
            return self._defaults[partname.ext]
        tmpl = "no content type for partname '%s' in [Content_Types].xml"
        raise KeyError(tmpl % partname)

    @staticmethod
    def from_xml(content_types_xml):
        """Return a new |_ContentTypeMap| instance populated with the contents of
        `content_types_xml`."""
        types_elm = parse_xml(content_types_xml)
        ct_map = _ContentTypeMap()
        for o in types_elm.overrides:
            ct_map._add_override(o.partname, o.content_type)
        for d in types_elm.defaults:
            ct_map._add_default(d.extension, d.content_type)
        return ct_map

    def _add_default(self, extension, content_type):
        """Add the default mapping of `extension` to `content_type` to this content type
        mapping."""
        self._defaults[extension] = content_type

    def _add_override(self, partname, content_type):
        """Add the default mapping of `partname` to `content_type` to this content type
        mapping."""
        self._overrides[partname] = content_type


class _SerializedPart:
    """Value object for an OPC package part.

    Provides access to the partname, content type, blob, and serialized relationships
    for the part.
    """

    def __init__(self, partname, content_type, reltype, blob, srels):
        super(_SerializedPart, self).__init__()
        self._partname = partname
        self._content_type = content_type
        self._reltype = reltype
        self._blob = blob
        self._srels = srels

    @property
    def partname(self):
        return self._partname

    @property
    def content_type(self):
        return self._content_type

    @property
    def blob(self):
        return self._blob

    @property
    def reltype(self):
        """The referring relationship type of this part."""
        return self._reltype

    @property
    def srels(self):
        return self._srels


class _SerializedRelationship:
    """Value object representing a serialized relationship in an OPC package.

    Serialized, in this case, means any target part is referred to via its partname
    rather than a direct link to an in-memory |Part| object.
    """

    def __init__(self, baseURI, rel_elm):
        super(_SerializedRelationship, self).__init__()
        self._baseURI = baseURI
        self._rId = rel_elm.rId
        self._reltype = rel_elm.reltype
        self._target_mode = rel_elm.target_mode
        self._target_ref = rel_elm.target_ref

    @property
    def is_external(self):
        """True if target_mode is ``RTM.EXTERNAL``"""
        return self._target_mode == RTM.EXTERNAL

    @property
    def reltype(self):
        """Relationship type, like ``RT.OFFICE_DOCUMENT``"""
        return self._reltype

    @property
    def rId(self):
        """Relationship id, like 'rId9', corresponds to the ``Id`` attribute on the
        ``CT_Relationship`` element."""
        return self._rId

    @property
    def target_mode(self):
        """String in ``TargetMode`` attribute of ``CT_Relationship`` element, one of
        ``RTM.INTERNAL`` or ``RTM.EXTERNAL``."""
        return self._target_mode

    @property
    def target_ref(self):
        """String in ``Target`` attribute of ``CT_Relationship`` element, a relative
        part reference for internal target mode or an arbitrary URI, e.g. an HTTP URL,
        for external target mode."""
        return self._target_ref

    @property
    def target_partname(self):
        """|PackURI| instance containing partname targeted by this relationship.

        Raises ``ValueError`` on reference if target_mode is ``'External'``. Use
        :attr:`target_mode` to check before referencing.
        """
        if self.is_external:
            msg = (
                "target_partname attribute on Relationship is undefined w"
                'here TargetMode == "External"'
            )
            raise ValueError(msg)
        # lazy-load _target_partname attribute
        if not hasattr(self, "_target_partname"):
            self._target_partname = PackURI.from_rel_ref(self._baseURI, self.target_ref)
        return self._target_partname


class _SerializedRelationships:
    """Read-only sequence of |_SerializedRelationship| instances corresponding to the
    relationships item XML passed to constructor."""

    def __init__(self):
        super(_SerializedRelationships, self).__init__()
        self._srels = []

    def __iter__(self):
        """Support iteration, e.g. 'for x in srels:'."""
        return self._srels.__iter__()

    @staticmethod
    def load_from_xml(baseURI, rels_item_xml):
        """Return |_SerializedRelationships| instance loaded with the relationships
        contained in `rels_item_xml`.

        Returns an empty collection if `rels_item_xml` is |None|.
        """
        srels = _SerializedRelationships()
        if rels_item_xml is not None:
            rels_elm = parse_xml(rels_item_xml)
            for rel_elm in rels_elm.Relationship_lst:
                srels._srels.append(_SerializedRelationship(baseURI, rel_elm))
        return srels



"""Custom element classes related to the numbering part."""

from docx.oxml.parser import OxmlElement
from docx.oxml.shared import CT_DecimalNumber
from docx.oxml.simpletypes import (
    ST_DecimalNumber,
    ST_LevelSuffix,
    ST_NumberFormat,
    ST_String,
    ST_MultiLevelType,
    ST_TwipsMeasure,
    ST_SignedTwipsMeasure,
    ST_OnOff,
    ST_LongHexNumber,
)
from docx.oxml.xmlchemy import (
    BaseOxmlElement,
    OneAndOnlyOne,
    RequiredAttribute,
    ZeroOrMore,
    ZeroOrOne,
    OptionalAttribute,
    Choice,
)


class CT_LevelText(BaseOxmlElement):
    """``<w:lvlText>`` element, which specifies
    the formatting of the numeral in a numbered
    list.
    """

    val = OptionalAttribute("w:val", ST_String)
    null = OptionalAttribute("w:null", ST_OnOff)

    @classmethod
    def new(cls, val):
        """
        Return a new ``<w:lvlText>`` element with
        ``val`` attribute set to *val*
        """
        lvlText = OxmlElement("w:lvlText")
        lvlText.val = val
        return lvlText


class CT_LevelSuffix(BaseOxmlElement):
    """
    ``<w:suff>`` element, which specifies the form of the space
    between a list number and the list paragraph
    """

    val = RequiredAttribute("w:val", ST_LevelSuffix)

    @classmethod
    def new(cls, val):
        """
        Return a new ``<w:suff>`` element with ``val``
        attribute set to *val*
        """
        suff = OxmlElement("w:suff")
        suff.val = val
        return suff


class CT_NumFmt(BaseOxmlElement):
    """
    ``<w:numFmt>`` element, which specifies the formatting
    of the numeral in a numbered list
    """

    val = RequiredAttribute("w:val", ST_NumberFormat)
    fmt = OptionalAttribute("w:format", ST_String)

    @classmethod
    def new(cls, val):
        """
        Return a new ``<w:numFmt>`` element with ``val``
        attrribute set to *val*
        """
        numFmt = OxmlElement("w:numFmt")
        numFmt.val = val
        return numFmt


class CT_LvlLegacy(BaseOxmlElement):
    """
    ``<w:legacy>`` element. Implemented here in
    case the module eventually supports parsing
    of documents in the target legacy format.
    """

    legacy = OptionalAttribute("w:legacy", ST_OnOff)
    legacySpace = OptionalAttribute("w:legacySpace", ST_TwipsMeasure)
    legacyIndent = OptionalAttribute("w:legacyIndent", ST_SignedTwipsMeasure)


class CT_MultiLevelType(BaseOxmlElement):
    """
    ``<w:multiLevelType>`` element, which indicates
    whether a numbering style is single-level,
    multi-level, or hybrid.
    """

    val = RequiredAttribute("w:val", ST_MultiLevelType)

    @classmethod
    def new(cls, val):
        """
        Return a new ``<w:multiLevelType>`` element with ``val``
        attribute set to *val*
        """
        multiLevelType = OxmlElement("w:multiLevelType")
        multiLevelType.val = val
        return multiLevelType


class CT_AbstractNum(BaseOxmlElement):
    """
    ``<w:abstractNum>`` element, which collects
    all of the level-specific style information
    for a particular style.
    """

    nsid = ZeroOrMore("w:nsid")
    multiLevelType = ZeroOrMore("w:multiLevelType")
    tmpl = ZeroOrMore("w:tmpl")
    name = ZeroOrMore("w:name")
    styleLink = ZeroOrMore("w:styleLink")
    numStyleLink = ZeroOrMore("w:numStyleLink")
    lvl = ZeroOrMore("w:lvl")
    abstractNumId = RequiredAttribute("w:abstractNumId", ST_DecimalNumber)

    @classmethod
    def new(cls, abstractNumId):
        """
        Return a new ``<w:abstractNum>`` element with ``abstractNumId``
        set to *abstractNumId*.
        """
        abstractNum = OxmlElement("w:abstractNum")
        abstractNum.abstractNumId = abstractNumId
        return abstractNum


class CT_Lvl(BaseOxmlElement):
    """
    ``<w:lvl>`` element, which contains all of
    the actual, level-specific formatting for
    a list style.
    """

    start = ZeroOrMore("w:start")
    numFmt = ZeroOrMore("w:numFmt")
    lvlRestart = ZeroOrMore("w:lvlRestart")
    pStyle = ZeroOrMore("w:pStyle")
    isLgl = ZeroOrMore("w:isLgl")
    suff = ZeroOrMore("w:suff")
    lvlText = ZeroOrMore("w:lvlText")
    lvlPicBulletId = ZeroOrMore("w:lvlPicBulletId")
    legacy = ZeroOrMore("w:legacy")
    lvlJc = ZeroOrMore("w:lvlJc")
    pPr = ZeroOrMore("w:pPr")
    rPr = ZeroOrMore("w:rPr")
    ilvl = RequiredAttribute("w:ilvl", ST_DecimalNumber)
    tplc = OptionalAttribute("w:tplc", ST_LongHexNumber)
    tentative = OptionalAttribute("w:tentative", ST_OnOff)

    @classmethod
    def new(cls, ilvl):
        """
        Return a new ``<w:lvl>`` element with ``ilvl``
        attribute set to *ilvl*
        """
        lvl = OxmlElement("w:lvl")
        lvl.ilvl = ilvl
        return lvl


class CT_NumPicBullet(BaseOxmlElement):
    """
    ``<w:numPicBullet>``` for specifying
    a picture or SVG drawing as the bullet
    symbol in a bulleted list.
    """

    pict = Choice("w:pict")
    drawing = Choice("w:drawing")
    numPicBulletId = RequiredAttribute("w:numPicBulletId", ST_DecimalNumber)

    @classmethod
    def new(cls, Id):
        """
        Return a new ``<w:numPicBullet>`` element with ``numPicBulletId``
        attribute set to *numPicBulletId*
        """
        numPicBullet = OxmlElement("w:numPicBullet")
        numPicBullet.numPicBulletId = Id
        return numPicBullet


class CT_Num(BaseOxmlElement):
    """``<w:num>`` element, which represents a concrete list definition instance, having
    a required child <w:abstractNumId> that references an abstract numbering definition
    that defines most of the formatting details."""

    abstractNumId = OneAndOnlyOne("w:abstractNumId")
    lvlOverride = ZeroOrMore("w:lvlOverride")
    numId = RequiredAttribute("w:numId", ST_DecimalNumber)

    def add_lvlOverride(self, ilvl):
        """Return a newly added CT_NumLvl (<w:lvlOverride>) element having its ``ilvl``
        attribute set to `ilvl`."""
        return self._add_lvlOverride(ilvl=ilvl)

    @classmethod
    def new(cls, num_id, abstractNum_id):
        """Return a new ``<w:num>`` element having numId of `num_id` and having a
        ``<w:abstractNumId>`` child with val attribute set to `abstractNum_id`."""
        num = OxmlElement("w:num")
        num.numId = num_id
        abstractNumId = CT_DecimalNumber.new("w:abstractNumId", abstractNum_id)
        num.append(abstractNumId)
        return num


class CT_NumLvl(BaseOxmlElement):
    """``<w:lvlOverride>`` element, which identifies a level in a list definition to
    override with settings it contains."""

    startOverride = ZeroOrOne("w:startOverride", successors=("w:lvl",))
    ilvl = RequiredAttribute("w:ilvl", ST_DecimalNumber)

    def add_startOverride(self, val):
        """Return a newly added CT_DecimalNumber element having tagname
        ``w:startOverride`` and ``val`` attribute set to `val`."""
        return self._add_startOverride(val=val)


class CT_NumPr(BaseOxmlElement):
    """A ``<w:numPr>`` element, a container for numbering properties applied to a
    paragraph."""

    ilvl = ZeroOrOne("w:ilvl", successors=("w:numId", "w:numberingChange", "w:ins"))
    numId = ZeroOrOne("w:numId", successors=("w:numberingChange", "w:ins"))

    # @ilvl.setter
    # def _set_ilvl(self, val):
    #     """
    #     Get or add a <w:ilvl> child and set its ``w:val`` attribute to `val`.
    #     """
    #     ilvl = self.get_or_add_ilvl()
    #     ilvl.val = val

    # @numId.setter
    # def numId(self, val):
    #     """
    #     Get or add a <w:numId> child and set its ``w:val`` attribute to
    #     `val`.
    #     """
    #     numId = self.get_or_add_numId()
    #     numId.val = val


class CT_Numbering(BaseOxmlElement):
    """``<w:numbering>`` element, the root element of a numbering part, i.e.
    numbering.xml."""

    num = ZeroOrMore("w:num", successors=("w:numIdMacAtCleanup",))
    abstractNum = ZeroOrMore("w:abstractNum")

    def add_num(self, abstractNum_id):
        """Return a newly added CT_Num (<w:num>) element referencing the abstract
        numbering definition identified by `abstractNum_id`."""
        next_num_id = self._next_numId
        num = CT_Num.new(next_num_id, abstractNum_id)
        return self._insert_num(num)

    def num_having_numId(self, numId):
        """Return the ``<w:num>`` child element having ``numId`` attribute matching
        `numId`."""
        xpath = './w:num[@w:numId="%d"]' % numId
        try:
            return self.xpath(xpath)[0]
        except IndexError:
            raise KeyError("no <w:num> element with numId %d" % numId)

    def abstractNum_having_abstractNumId(self, abstractNumId):
        """
        Return the ``<w:abstractNum>`` child element having ``abstractNumId`` attribute
        matching *numId*.
        """
        xpath = './w:abstractNum[@w:abstractNumId="%d"]' % abstractNumId
        try:
            return self.xpath(xpath)[0]
        except IndexError:
            raise KeyError("no <w:abstractNum> element with abstractNumId %d" % abstractNumId)

    @property
    def _next_numId(self):
        """The first ``numId`` unused by a ``<w:num>`` element, starting at 1 and
        filling any gaps in numbering between existing ``<w:num>`` elements."""
        numId_strs = self.xpath("./w:num/@w:numId")
        num_ids = [int(numId_str) for numId_str in numId_strs]
        for num in range(1, len(num_ids) + 2):
            if num not in num_ids:
                break
        return num
